from pathlib import Path
import subprocess

from hardware_test_tools.UaDut import UaDut
from hardware_test_tools.UaDfuApp import UaDfuApp


def create_dfu_bin(fw_path, bin_path, upgrade_num):
    cmd = [
        "xflash",
        "--factory-version",
        "15.3",
        "--upgrade",
        f"{upgrade_num}",
        fw_path,
        "-o",
        bin_path,
    ]
    ret = subprocess.run(cmd, capture_output=True, timeout=10, text=True)
    assert (
        ret.returncode == 0
    ), f"Failed to generate DFU binary for {fw_path}:\n{ret.stdout + ret.stderr}"


def test_dfu(pytestconfig):
    xtag_id = pytestconfig.getoption("xtag_id")
    assert xtag_id, "--xtag-id option must be provided on the command line"

    fw_path = (
        Path(__file__).parents[1] / "app_an02016" / "bin" / "UA" / "app_an02016_UA.xe"
    )
    assert fw_path.exists(), f"Firmware not present at {fw_path}"

    fw_dir = Path(__file__).parents[1] / "app_an02016" / "bin"
    fw_factory = fw_dir / "UA" / "app_an02016_UA.xe"

    # Generate DFU binaries for the upgrade firmware versions
    fw_upgrade0 = fw_dir / "upgrade0" / "app_an02016_upgrade0.xe"
    dfu_bin_upgrade0 = fw_dir / "upgrade0" / "app_an02016_upgrade0.bin"
    create_dfu_bin(fw_upgrade0, dfu_bin_upgrade0, 1)

    fw_upgrade1 = fw_dir / "upgrade1" / "app_an02016_upgrade1.xe"
    dfu_bin_upgrade1 = fw_dir / "upgrade1" / "app_an02016_upgrade1.bin"
    create_dfu_bin(fw_upgrade1, dfu_bin_upgrade1, 2)

    pid = 0x18
    in_chans = 2
    out_chans = 2
    prod_str = "XMOS xCORE.ai AWE (UAC2.0)"

    with UaDut(xtag_id, fw_factory, pid, prod_str, in_chans, out_chans, xflash=True) as dut:
        dfu_test = UaDfuApp(dut.driver_guid, pid)

        version_factory = dfu_test.get_bcd_version()

        # perform DFU from factory to upgrade partition
        dfu_test.download(dfu_bin_upgrade0)
        version_upgrade0 = dfu_test.get_bcd_version()
        assert version_upgrade0 != version_factory

        # perform DFU from upgrade to upgrade partition
        dfu_test.download(dfu_bin_upgrade1)
        version_upgrade1 = dfu_test.get_bcd_version()
        assert version_upgrade1 not in [version_factory, version_upgrade0]

        # revert to the factory version
        dfu_test.revert_factory()
        version_revert = dfu_test.get_bcd_version()
        assert version_revert == version_factory
