import configparser
import subprocess
import platform
from pathlib import Path
import time
import os
import shutil

import mido
import sounddevice as sd

from hardware_test_tools.XcoreApp import XcoreApp


class UaDut(XcoreApp):
    def __init__(
        self,
        adapter_id,
        fw_path,
        pid,
        prod_str,
        in_chans,
        out_chans,
        vid=0x20B1,
        winbuiltin=False,
        timeout=60,
        xflash=False,
    ):
        assert fw_path.exists(), f"Firmware not present at {fw_path}"

        self.clock_src = "Internal"
        self.pid = pid
        self.chan_i = in_chans
        self.chan_o = out_chans

        self.winbuiltin = winbuiltin
        if platform.system() == "Windows" and not self.winbuiltin:
            ini_path = (
                Path(os.environ["PROGRAMFILES"])
                / "XMOS"
                / "USB Audio Device Driver"
                / "x64"
                / "custom.ini"
            )
            assert (
                ini_path.exists()
            ), f"tusbaudio SDK custom.ini not found in expected location: {ini_path}"

            with open(ini_path, "r") as f:
                config = configparser.ConfigParser()
                config.read_file(f)
            try:
                self.driver_guid = config.get("DriverInterface", "InterfaceGUID")
            except (configparser.NoSectionError, configparser.NoOptionError):
                assert 0, f"Could not find InterfaceGUID in {ini_path}"
        else:
            self.driver_guid = None

        volcontrol_dir = Path(__file__).parents[1] / "volcontrol"
        if platform.system() == "Windows":
            volcontrol_path = volcontrol_dir / "x64" / "Release" / "volcontrol.exe"
        else:
            volcontrol_path = volcontrol_dir / "build" / "volcontrol"

        if volcontrol_path.exists():
            self.volcontrol_cmd = [volcontrol_path]
            if platform.system() == "Windows" and self.driver_guid:
                self.volcontrol_cmd.append(f"-g{self.driver_guid}")
        else:
            self.volcontrol_cmd = None

        self.prod_str = prod_str
        self.dev_name = "ASIO4ALL v2" if self.winbuiltin else prod_str

        super().__init__(fw_path, adapter_id, timeout=timeout, xflash=xflash)

    def wait_for_portaudio(self):
        for _ in range(10):
            time.sleep(1)

            # sounddevice must be terminated and re-initialised to get updated device info
            sd._terminate()
            sd._initialize()
            if any(self.prod_str in sd_dev["name"] for sd_dev in sd.query_devices()):
                return

        fail_str = f"Device ({self.prod_str}) failed to enumerate\n"

        # Device doesn't appear to have started, so dump the state of the xcore
        ret = subprocess.run(
            ["xrun", "--adapter-id", self.adapter_id, "--dump-state", self.xe_path],
            text=True,
            timeout=10,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )
        fail_str += "Register and thread state of the xcore device\n"
        fail_str += "\n".join(ret.stdout.splitlines())
        assert 0, fail_str

    def __enter__(self):
        super().__enter__()

        if platform.system() == "Windows":
            # Delay to allow the Windows Audio service to set up the new device
            time.sleep(40)

        self.wait_for_portaudio()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.volcontrol_cmd and self.clock_src != "Internal":
            # Reset the clock source to default before stopping the application
            self.set_clock_src("Internal")

        super().__exit__(exc_type, exc_val, exc_tb)

        if platform.system() == "Windows":
            # Delay to allow the Windows Audio service to settle after the device is removed
            time.sleep(15)

        # If usbdeview program is present on Windows, uninstall the device to force
        # re-enumeration next time and avoid caching of device features by the OS
        if platform.system() == "Windows":
            usbdeview_path = shutil.which("usbdeview")
            if usbdeview_path:
                subprocess.run(
                    [
                        usbdeview_path,
                        "/RunAsAdmin",
                        "/remove_by_pid",
                        f"0x20b1;{hex(self.pid)}",
                    ],
                    timeout=10,
                )

    def _set_stream_format(self, direction, samp_freq, num_chans, bit_depth, fail_on_err):
        cmd = self.volcontrol_cmd + [
            "--set-format",
            direction,
            f"{samp_freq}",
            f"{num_chans}",
            f"{bit_depth}",
        ]
        ret = subprocess.run(cmd, timeout=30, capture_output=True, text=True)
        if fail_on_err:
            assert ret.returncode == 0, f"failed to setup stream format: {direction}, {samp_freq} fs, {num_chans} channels, {bit_depth} bit\n{ret.stdout}\n{ret.stderr}"

        return ret.returncode

    def set_stream_format(self, direction, samp_freq, num_chans, bit_depth, fail_on_err=True):
        assert self.volcontrol_cmd

        if self.winbuiltin:
            # Cannot change the stream format
            return

        # Set the stream format that the user has asked for
        return self._set_stream_format(direction, samp_freq, num_chans, bit_depth, fail_on_err)

    def set_clock_src(self, clock_src):
        assert clock_src in [
            "Internal",
            "SPDIF",
            "ADAT",
        ], f"Invalid clock source: {clock_src}"
        assert self.volcontrol_cmd

        cmd = self.volcontrol_cmd + ["--clock", clock_src]
        ret = subprocess.run(cmd, capture_output=True, text=True, timeout=30)
        # Sometimes a valid clock isn't seen, so print a message instead of asserting until sw_usb_audio issue #255 is resolved
        #assert (
        #    ret.returncode == 0
        #), f"setting clock source failed, cmd: {cmd}\nstdout:\n{ret.stdout}\nstderr:\n{ret.stderr}"
        if ret.returncode != 0:
            print(f"setting clock source failed, cmd: {cmd}\nstdout:\n{ret.stdout}\nstderr:\n{ret.stderr}")

        self.clock_src = clock_src

    def volume_reset(self):
        assert (
            self.chan_i == self.chan_o
        ), "Volume reset is only supported by volcontrol for applications with the same number of channels for input and output"
        assert self.volcontrol_cmd

        num_chans = self.chan_i
        cmd = self.volcontrol_cmd + ["--resetall", f"{num_chans + 1}"]
        ret = subprocess.run(cmd, capture_output=True, text=True, timeout=30)
        assert (
            ret.returncode == 0
        ), f"resetting volume failed, cmd: {cmd}\nstdout:\n{ret.stdout}\nstderr:\n{ret.stderr}"

    def volume_set(self, direction, channel, value):
        assert direction in ["input", "output"]
        assert self.volcontrol_cmd

        num_chans = self.chan_i if direction == "input" else self.chan_o
        # volcontrol takes a value one larger than the actual number of channels
        num_chans += 1

        cmd = self.volcontrol_cmd + ["--set", direction, f"{channel + 1}", f"{value}"]
        ret = subprocess.run(cmd, capture_output=True, text=True, timeout=30)
        assert (
            ret.returncode == 0
        ), f"setting channel volume failed, cmd: {cmd}\nstdout:\n{ret.stdout}\nstderr:\n{ret.stderr}"

    def volume_set_master(self, direction, value):
        assert direction in ["input", "output"]
        assert self.volcontrol_cmd

        num_chans = self.chan_i if direction == "input" else self.chan_o
        # volcontrol takes a value one larger than the actual number of channels
        num_chans += 1

        cmd = self.volcontrol_cmd + ["--set", direction, "0", f"{value}"]
        ret = subprocess.run(cmd, capture_output=True, text=True, timeout=30)
        assert (
            ret.returncode == 0
        ), f"setting master volume failed, cmd: {cmd}\nstdout:\n{ret.stdout}\nstderr:\n{ret.stderr}"

    def wait_for_midi_ports(
        self, timeout_s=60, restart_attempts=15
    ):  # Win 11 agent has been seen to take up to 40s
        for restart_count in range(restart_attempts):
            for i in range(timeout_s):
                midi_in = [md for md in mido.get_input_names() if "XMOS" in md]
                midi_out = [md for md in mido.get_output_names() if "XMOS" in md]
                if len(midi_in) > 0 and len(midi_out) > 0:
                    self.midi_in = midi_in[0]
                    self.midi_out = midi_out[0]
                    print(f"Hooray! XMOS MIDI ports found: {self.midi_in}, {self.midi_out}")
                    return True
                time.sleep(1)
                print(f"MIDI ports not found... retrying {i+1} of {timeout_s}")

            print(f"No XMOS MIDI ports found: {mido.get_input_names()}, {mido.get_output_names()}")

            # Stop and restart the device
            self.__exit__(None, None, None)
            self.__enter__()
            return False
